// automatic selection of pid parameters

function autoSelectTemplate(zoneID, force)
{
    var eventSend = false;
    // parameter validation - ignore pseudo zone 0
    if ((zoneID == null) || (zoneID == '0')) {
        lWarn.logln('PID autoselection skipped: parameter validation - ignore pseudo zone 0');
        return eventSend;
    }

    // an un-configured zone is per default in mode 'autoselect'
    var pZone = Property.getNode('controllerConfig/' + zoneID);
    if (pZone == null) {
        storeParameter('controllerConfig/' + zoneID + '/zoneID', zoneID);
        storeParameter('controllerConfig/' + zoneID + '/definitionID', -1);
        storeParameter('controllerConfig/' + zoneID + '/autoselect', 1);
        Property.store();
    }

    // autoselect active for this zone?
    var autoselect = Property.getProperty('controllerConfig/' + zoneID + '/autoselect');
    if ((autoselect == null) || (autoselect <= 0)) {
        lWarn.logln('PID autoselection skipped: autoselection not enabled for zone ' + zoneID + ': ' + autoselect);
        return eventSend;
    }

    // definitionID is the numerical index of the controller parameter template
    // maybe negative, -1 => nothing selected, -2 => inconstistent valve types in zone, -3 => all valves Unknown
    var definitionID = Property.getProperty('controllerConfig/' + zoneID + '/definitionID');
    var zoneDefinitionID = DefinitionIDs.AutoNothingSelected;
    if (definitionID != null && definitionID >= 0) {
        zoneDefinitionID = definitionID;
    }

    var zoneValveType = "";
    var allValvesUnknown = true;
    var selectionStatus = 0;
    var numValveDevices = 0;

    lInfo.logln('PID autoselection algorithm running for zone ' + zoneID);
    var zone = getZoneByID(parseInt(zoneID));
    var devices = zone.getDevices();
    devices.perform(function (device) {

        // inactive devices must be ignored
        if (device.isPresent <= 0) {
            return;
        }

        // device valve type property
        var valveProperty = null;
        try {
            valveProperty = device.getPropertyNode().getChild('ValveType');
        } catch (e) {
            return;
        }
        if (valveProperty == null) {
            return;
        }
        var valveType = valveProperty.getValue();
        lInfo.logln('PID autoselection device: ' + device.dsuid + ' has valve type: ' + valveType);

        // run template selection algorithm
        if (valveType != null && valveType != "") {
            numValveDevices ++;
        }
        if (valveType != null && valveType !== "Unknown") {
            allValvesUnknown = false;
            if (valveType == "Floor" || valveType == "Wall") {
                valveType = "FloorWall";
            }
            if (selectionStatus == 0) {
                // nothing else found so far, the first device setting is saved
                zoneValveType = valveType;
                selectionStatus = 1;
            }
            else if (selectionStatus == 1) {
                // inconsistent/different types?
                if (zoneValveType != valveType) {
                    zoneValveType = "ERROR";
                    selectionStatus = -1;
                }
            }
        }
    });

    // does this zone have a temperature sensor? without sensor no pid control possible...
    var sensorDevice = zone.getAssignedSensor(9);
    if (sensorDevice == null || !sensorDevice.dsuid) {
        lWarn.logln('No sensors in zone ' + zoneID + ': reset controller');
        var statusValue = DefinitionIDs.AutoNothingSelected;

        var pDefinitionID = Property.getNode('controllerConfig/' + zoneID + '/definitionID');
        pDefinitionID.setValue(statusValue);
        Property.store();

        new Event('heating-controller.controller-config', {
            actions : 'resetControllerMode',
            zoneID : '' + zoneID,
            autoselect: 1,
            definitionID: statusValue
        }).raise();

        return true;
    }

    var result = '';
    result += '/ valve result: ' + zoneValveType + ' /';
    result += '/ selection status: ' +  selectionStatus + ' /';
    if (allValvesUnknown) {
        result += '/ all valves unknown /';
    }
    lInfo.logln('PID autoselection result for zone ' + zoneID + ': ' + result);

    // if we found consistent valve types try to find a best-match in the configuration templates,
    // the controllerDefinitions/ have been loaded from the cloud on startup
    if (selectionStatus > 0) {
        var pControllerTemplates = Property.getNode('controllerDefinitions').getChildren();
        var n;
        var newDefinitionID = -1;
        for (n = 0; n < pControllerTemplates.length; n ++) {
            var templateType = pControllerTemplates[n].getChild('type').getValue();
            if (zoneValveType == 'Radiator' && templateType == 1) {
                newDefinitionID = pControllerTemplates[n].getChild('id').getValue();
                break;
            }
            if (zoneValveType == 'FloorWall' && templateType == 2) {
                newDefinitionID = pControllerTemplates[n].getChild('id').getValue();
                break;
            }
        }
        lInfo.logln('PID autoselection template id for zone ' + zoneID + ': old id ' + zoneDefinitionID + ' => ' + newDefinitionID);
        if (newDefinitionID < 0) {
            lWarn.logln('PID autoselection template not found for zone ' + zoneID + ' and heating type ' + zoneValveType);
        }
        else if ((force == false) && (newDefinitionID == zoneDefinitionID)) {
            lInfo.logln('PID autoselection => nothing to change');
        }
        else {
            lInfo.logln('PID autoselection set controller config');
            new Event('heating-controller.controller-config', {
                actions : 'setControllerModePID',
                zoneID : '' + zoneID,
                definitionID: '' + newDefinitionID,
                autoselect: 1
            }).raise();
            eventSend = true;
        }
    }

    // if not consistent disable a possibly running controller and save the status AutoSelectionFailed (used by the UI)
    if (selectionStatus == -1) {
        lInfo.logln('PID autoselection failure for zone ' + zoneID);
        var pDefinitionID = Property.getNode('controllerConfig/' + zoneID + '/definitionID');
        pDefinitionID.setValue(DefinitionIDs.AutoSelectionFailed);
        Property.store();

        new Event('heating-controller.controller-config', {
            actions : 'resetControllerMode',
            zoneID : '' + zoneID,
            autoselect: 1,
            definitionID: DefinitionIDs.AutoSelectionFailed
        }).raise();
        eventSend = true;
    }

    // if no selection possible save the status AutoNothingSelected or AutoSelectionPossible accordingly (used by the UI)
    if ((selectionStatus == 0) || allValvesUnknown) {
        var statusValue = DefinitionIDs.ModeOff;
        if ((numValveDevices > 0) && allValvesUnknown) {
            statusValue = DefinitionIDs.AutoSelectionPossible;
        }

        var pDefinitionID = Property.getNode('controllerConfig/' + zoneID + '/definitionID');
        pDefinitionID.setValue(statusValue);
        Property.store();

        new Event('heating-controller.controller-config', {
            actions : 'resetControllerMode',
            zoneID : '' + zoneID,
            autoselect: 1,
            definitionID: statusValue
        }).raise();
        eventSend = true;
    }
    return eventSend;
}

//------------------------------------------------------------------------------
// getHeatingControllerDsmForZone get the controller dsuid
//------------------------------------------------------------------------------
function getHeatingControllerDsmForZone(zoneID) {
    var zZone = getZoneByID(parseInt('' + zoneID, 10));
    var config = zZone.getTemperatureControlConfiguration();
    return config.ControlDSUID;
}

//------------------------------------------------------------------------------
// evaluateByDSUID get meter by lowest dsuid
// input meterList
// returns dsuid of meter with lowest dsuid
//------------------------------------------------------------------------------
function evaluateByDSUID(meterList) {
    if (meterList.length > 1) {
        var oMeter = meterList[0];
        for (var iIndex = 1; iIndex < meterList.length; iIndex++) {
            if (oMeter < meterList[iIndex]) {
                oMeter = meterList[iIndex];
            }
        }
        return oMeter;
    }
    return meterList[0];
}

//------------------------------------------------------------------------------
// getDSMListByType get a list of dSM by given type and zone
//------------------------------------------------------------------------------
function getDSMListByType(dSMType, zoneID) {
    var oPossibleCircuits = [];
    // get all dSM of type dSMType
    var odSMArray = Property.getNode('/apartment/dSMeters').getChildren();
    for ( var iIndex = 0; iIndex < odSMArray.length; iIndex++) {
        //lInfo.logln('search for ' + odSMArray[iIndex].getPath() + '/busMemberType');
        if (dSMType == Property.getProperty(odSMArray[iIndex].getPath() + '/busMemberType')) {
            //lInfo.logln('found dSM with busMemberType ' + odSMArray[iIndex].getChild('dSUID').getValue());
            oPossibleCircuits.push(odSMArray[iIndex].getChild('dSUID').getValue());
        }
    }
    // Select the dSM with zone
    var pidCircuits = [];
    for (var iIndex = 0; iIndex < oPossibleCircuits.length; iIndex++) {
        var dSM = oPossibleCircuits[iIndex];
//        lInfo.logln('looking at dSM: ' + dSM);
        if (Property.getNode('/apartment/dSMeters/' + dSM + '/zones') != null) {
//            lInfo.logln('dSM ' + dSM + ' has zone node ...');
            var oDSMZoneNode = Property.getNode('/apartment/dSMeters/' + dSM + '/zones').getChildren();
            lInfo.logln('dSM ' + dSM + ' has zone node count ' + oDSMZoneNode.length);
            for (var jIndex = 0; jIndex < oDSMZoneNode.length; jIndex++) {
//                lInfo.logln('check dSM ' + dSM + ' if zone ' + zoneID + ' is registered');
                if (zoneID == oDSMZoneNode[jIndex].getChild('ZoneID').getValue()) {
                    pidCircuits.push(dSM);
                }
            }
        } else {
//            lInfo.logln('no zones on meter ' + dSM);
        }
    }
    return pidCircuits;
}

//------------------------------------------------------------------------------
// move the controller from dSMOld to dSMNew for the given zoneID
//------------------------------------------------------------------------------
function migrateController(dSMOld, dSMNew, zoneID, force) {
    lInfo.logln('migrateController from: ' + dSMOld + ' to: ' + dSMNew + ' for zone: ' + zoneID);

    // parameter validation - ignore pseudo zone 0
    if ((zoneID == null) || (zoneID == '0')) {
        lWarn.logln('PID migrateController skipped: parameter validation - ignore pseudo zone 0');
        return;
    }

    if ((force == false) && (dSMOld == dSMNew)) {
        lInfo.logln('migrateController migration not needed');
        return;
    }

    var zZone = getZoneByID(parseInt('' + zoneID, 10));

    // get parameter from old DSM
    var oldConfig = zZone.getTemperatureControlConfiguration();

    // stop controller on old DSM
    if (dSMOld != null) {
        zZone.setTemperatureControlConfiguration(dSMOld, { ControlMode : 0 });
        zZone.setTemperatureControlConfiguration("0000000000000000000000000000000000");
    }

    // assign to new controller
    if (dSMNew != null) {
        zZone.setTemperatureControlConfiguration(dSMNew, oldConfig);
    }
}

//------------------------------------------------------------------------------
// evaluate the type of the dSM. returns the device type.
// dSM the dsuid of the dSM
//------------------------------------------------------------------------------
function getTypeForDsm(dSM) {
    var type = Property.getNode('/apartment/dSMeters/' + dSM + '/busMemberType').getValue();
    return type;
}

//------------------------------------------------------------------------------
// get a list of heating actuators for the given dSM and zone
// dSM dsuid of the dSM
// zoneID  the zone ID
// returns a list with dsuid of actuators
//------------------------------------------------------------------------------
function getHeatingActors(dSM, zoneID) {
    var odSMArray = Property.getNode('/apartment/zones/zone' + zoneID + '/devices').getChildren();
    var oHeatingCircuits = [];
    for (var iIndex = 0; iIndex < odSMArray.length; iIndex++) {
        // lInfo.logln('search for ' + odSMArray[iIndex].getPath() + '/isValveType');
        if (Property.getProperty(odSMArray[iIndex].getPath() + '/isValveType')) {
            // lInfo.logln('search for ' + odSMArray[iIndex].getPath() + '/ZoneID');
            if (zoneID == Property.getProperty(odSMArray[iIndex].getPath() + '/ZoneID')) {
                // lInfo.logln('search for ' + odSMArray[iIndex].getPath() + '/DSMeterDSUID');
                if (dSM == Property.getProperty(odSMArray[iIndex].getPath() + '/DSMeterDSUID')) {
                    // lInfo.logln('found device of type heating valve  ' + odSMArray[iIndex].getChild('dSUID').getValue());
                    oHeatingCircuits.push(odSMArray[iIndex].getChild('dSUID').getValue());
                }
            }
        }
    }
    return oHeatingCircuits;
}

//------------------------------------------------------------------------------
// create a list with attribute 'meter' and attribute 'number of heating actors'
// input dSMList list of dSM's
// input zoneID  the zone ID
// returns an array with a structure [meter (dsuid), count (actuators)]
//------------------------------------------------------------------------------
function createHeatingActorsList(dSMList, zoneID) {
    // evaluate number of actors per meter
    var meterActuator = new Array();
    for (var iIndex= 0; iIndex < dSMList.length; iIndex++) {
        var count = getHeatingActors(dSMList[iIndex], zoneID).length;
        meterActuator.push({meter: dSMList[iIndex], count: count});
        lInfo.logln('Heating Actors for dSM: ' + dSMList[iIndex] +
                    ' zone: ' + zoneID +
                    ' count: ' + count);
    };
    return meterActuator;
}

//------------------------------------------------------------------------------
// evaluate maxima: returns a list with [meters(dsuid)]
// input meterActuatorList a list [meter (dsuid), count (actuators)]
// returns an array [meters(dsuid)]
//------------------------------------------------------------------------------
function evaluateMaxima(meterActuatorList) {
    var max = 0;
    for (var iIndex = 0; iIndex < meterActuatorList.length; iIndex++) {
        if (max < meterActuatorList[iIndex].count) {
            max = meterActuatorList[iIndex].count;
        }
    }
    //lInfo.logln('evaluateMaxima max value: ' + max);
    var oMeters = [];
    for (var iIndex = 0; iIndex < meterActuatorList.length; iIndex++) {
        if (max == meterActuatorList[iIndex].count) {
            oMeters.push(meterActuatorList[iIndex].meter);
            lInfo.logln('evaluateMaxima: Device is: ' +
                        meterActuatorList[iIndex].meter);
        }
    }
    return oMeters;
}

//------------------------------------------------------------------------------
// automatic selection of dSM
//------------------------------------------------------------------------------
function autoSelectDsm(zoneID, force) {
    // parameter validation - ignore pseudo zone 0
    if ((zoneID == null) || (zoneID == '0')) {
        lWarn.logln('autoSelectDsm of Dsm skipped: parameter validation - ignore pseudo zone 0');
        return;
    }

    // get assigned dSM for zone
    var currentdSM  = getHeatingControllerDsmForZone(zoneID);
    var currentType = BusMember_vDC;
    var currentHtDevCount = 0;

    if (currentdSM == null) {
        lWarn.logln('autoSelectDsm: no heating controller assigned for zone: ' + zoneID);
    } else {
        currentType = getTypeForDsm(currentdSM);
        currentHtDevCount = getHeatingActors(currentdSM, zoneID).length;
        lInfo.logln ('autoSelectDsm current dSM:' + currentdSM +
            ' type: ' + currentType +
            ' number of actors: ' + currentHtDevCount);
    }

    // are there any other DSM
    var dSMList = getDSMListByType(BusMember_dSM12, zoneID);
    if (dSMList.length > 0) {
        // get numbers of heating actuators per meter and evaluate maxima list
        var meterActuator = createHeatingActorsList(dSMList, zoneID);
        var maxMeterList = evaluateMaxima(meterActuator);
        if (maxMeterList.length > 1) {
            var sortMeter = evaluateByDSUID(maxMeterList);
            lInfo.logln ('migrate dSM to other dSM selected by dsuid');
            migrateController(currentdSM, sortMeter, zoneID, force);
        } else {
            migrateController(currentdSM, maxMeterList[0], zoneID, force);
        }
    } else {
        // stay on vdSM's
        var vdSMList = getDSMListByType(BusMember_vDC, zoneID);
        // get numbers of heating actuators per meter and evaluate maxima list
        var meterActuator = createHeatingActorsList(vdSMList, zoneID);
        var maxMeterList = evaluateMaxima(meterActuator);
        if (maxMeterList.length > 1) {
            var sortMeter = evaluateByDSUID(maxMeterList);
            lInfo.logln ('migrate vdSM to other vdSM by dsuid');
            migrateController(currentdSM, sortMeter, zoneID, force);
        } else {
            lInfo.logln ('migrate vdSM to other vdSM by maxima');
            migrateController(currentdSM, maxMeterList[0], zoneID, force);
        }
    }
}


if ((raisedEvent.name == 'heating-controller.autoselect') ||
    (raisedEvent.name == 'DeviceHeatingTypeChanged')) {

    try {
        autoSelectDsm(raisedEvent.parameter.zoneID, false);
    } catch (e) {
        lWarn.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }

    try {
        var eventSend = autoSelectTemplate(raisedEvent.parameter.zoneID, false);
        if (!eventSend) {
            new Event('heating-controller.answer', {
                          'requestingEvent' : raisedEvent.name,
                          'result' : -1
                      }).raise();
        }
    } catch (e) {
        lWarn.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }
}

else if (raisedEvent.name == 'DeviceEvent') {
    lInfo.logln ('DeviceEvent action: ' + raisedEvent.parameter.action + ' device ' + raisedEvent.source.dsuid);

    if (raisedEvent.parameter.action == 'moved') {
        try {
            autoSelectDsm(raisedEvent.parameter.oldZoneID, false);
            autoSelectTemplate(raisedEvent.parameter.oldZoneID, false);
        } catch (e) {
            lInfo.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
        }

        try {
            autoSelectDsm(raisedEvent.parameter.newZoneID, false);
            autoSelectTemplate(raisedEvent.parameter.newZoneID, false);
        } catch (e) {
            lInfo.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
        }
    }

    if (raisedEvent.parameter.action == 'added') {
        sendHeatingSeason();
    }

    if ((raisedEvent.parameter.action == 'deleted') ||
        (raisedEvent.parameter.action == 'added')) {
        autoSelectDsm(raisedEvent.source.zoneID, false);
        autoSelectTemplate(raisedEvent.source.zoneID, false);
    }
}

else if (raisedEvent.name == 'dsMeter_ready') {
    lInfo.logln ('dsMeter_ready');

    new Event('heating-controller.model_ready', {
        'time' : '+60',
        'unique' : 'Yes'
    }).raise();
}

else if (raisedEvent.name == 'heating-controller.model_ready') {
    lInfo.logln ('heating-controller.model_ready event');
    try {
        var zones = getZones();
        var index, len;
        for (index = 0, len = zones.length; index < len; ++index) {
            var zone = zones[index];
            lInfo.logln ('run autoSelectDsm() for zone ' + zone.name + '(' + zone.id + ')');
            autoSelectDsm(zone.id, false);
            autoSelectTemplate(zone.id, false);
        }
    } catch (e) {
        lWarn.logln('PID autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }
}

else if (raisedEvent.name == 'UpdateAutoselect') {
    lInfo.logln ('Update auto select triggered');
    try {
        var meterDSUID = raisedEvent.parameter.meterDSUID;
        var zones = getZones();
        var index, len;
        for (index = 0, len = zones.length; index < len; ++index) {
            var zone = zones[index];
            var heatConfig = zone.getTemperatureControlConfiguration();
            if (meterDSUID == heatConfig.ControlDSUID) {
                lInfo.logln ('UpdateAutoselect force run autoSelectDsm() for zone ' + zone.name + '(' + zone.id + ') on  meter (' + meterDSUID + ')');
                autoSelectDsm(zone.id, true);
                autoSelectTemplate(zone.id, true);
            }
        }
    } catch (e) {
        lWarn.logln('Update autoselection: error on event ' + raisedEvent.name + ': ' + e);
    }
}
